package org.uoc.petrel;

import java.util.ArrayList;
import java.util.List;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.util.Log;

public class PetrelDBAdapter {
	private static class DatabaseHelper extends SQLiteOpenHelper {

		DatabaseHelper(Context context) {
			super(context, DATABASE_NAME, null, DATABASE_VERSION);
		}

		@Override
		public void onCreate(SQLiteDatabase db) {

			db.execSQL(TABLE_PORTFOLIO_CREATE);
			db.execSQL(TABLE_ASSET_CREATE);
			db.execSQL(TABLE_TICKER_CREATE);
			db.execSQL(TABLE_PORTFOLIO_TO_ASSET_CREATE);
		}

		@Override
		public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
			Log.w(TAG, "Upgrading database from version " + oldVersion + " to "
					+ newVersion + ", which will destroy all old data");
			db.execSQL("DROP TABLE IF EXISTS portfolio");
			db.execSQL("DROP TABLE IF EXISTS asset");
			db.execSQL("DROP TABLE IF EXISTS ticker");
			onCreate(db);
		}
	}

	// Table Names
	public static final String TABLE_PORTFOLIO = "portfolio";
	public static final String TABLE_ASSET = "asset";
	public static final String TABLE_TICKER = "ticker";
	public static final String TABLE_PORTFOLIO_TO_ASSET = "portfolio_to_asset";

	// Key Names
	public static final String KEY_PORTFOLIO_ID = "_id";
	public static final String KEY_PORTFOLIO_NAME = "name";
	public static final String KEY_PORTFOLIO_MU = "mu";
	public static final String KEY_PORTFOLIO_SIGMA = "sigma";
	public static final String KEY_PORTFOLIO_RHO = "rho";

	private String[] PortfolioAllColumns = { KEY_PORTFOLIO_ID,
			KEY_PORTFOLIO_NAME, KEY_PORTFOLIO_MU, KEY_PORTFOLIO_SIGMA,
			KEY_PORTFOLIO_RHO };

	public static final String KEY_ASSET_ID = "_id";
	public static final String KEY_ASSET_NAME = "name";
	public static final String KEY_ASSET_W = "w";
	public static final String KEY_ASSET_MU = "mu";
	public static final String KEY_ASSET_SIGMA = "sigma";

	private String[] AssetAllColumns = { KEY_ASSET_ID, KEY_ASSET_NAME,
			KEY_ASSET_W, KEY_ASSET_MU, KEY_ASSET_SIGMA };

	public static final String KEY_TICKER_ID = "_id";
	public static final String KEY_TICKER_NAME = "name";
	public static final String KEY_TICKER_COMPANY = "company";

	public String[] TickerAllColumns = { KEY_TICKER_ID, KEY_TICKER_NAME,
			KEY_TICKER_COMPANY };

	public static final String KEY_PORTFOLIO_TO_ASSET_PORTFOLIO = "portfolio";
	public static final String KEY_PORTFOLIO_TO_ASSET_ASSET = "asset";

	private String[] PortfolioToAssetAllColumns = {
			KEY_PORTFOLIO_TO_ASSET_PORTFOLIO, KEY_PORTFOLIO_TO_ASSET_ASSET };

	// Database creation
	public static final String TABLE_TICKER_CREATE = "create table ticker (_id integer primary key autoincrement, "
			+ "name text not null, company text not null);";
	public static final String TABLE_ASSET_CREATE = "create table asset (_id integer primary key autoincrement, "
			+ "name text not null, w double, mu double, sigma double);";
	public static final String TABLE_PORTFOLIO_CREATE = "create table portfolio (_id integer primary key autoincrement, "
			+ "name text not null, mu real, sigma, real, rho real);";
	public static final String TABLE_PORTFOLIO_TO_ASSET_CREATE = "create table portfolio_to_asset (_id integer primary key autoincrement, "
			+ "portfolio integer reference portfolio, asset integer reference asset);";

	// Other constants
	private static final String DATABASE_NAME = "Petrel";
	private static final int DATABASE_VERSION = 2;
	private static final String TAG = "ErelisDbAdapter";

	// Members
	private DatabaseHelper mDbHelper;
	private SQLiteDatabase mDb;
	private final Context mCtx;
	
	public PetrelDBAdapter(Context ctx) {
		this.mCtx = ctx;
	}

	public PetrelDBAdapter open() throws SQLException {
		mDbHelper = new DatabaseHelper(mCtx);
		mDb = mDbHelper.getWritableDatabase();
		return this;
	}

	public void close() {
		mDbHelper.close();
	}

	//
	// Ticker methods
	//
	public Ticker createTicker(String name, String company) {
		ContentValues values = new ContentValues();
		values.put(KEY_TICKER_NAME, name);
		values.put(KEY_TICKER_COMPANY, company);
		long insertId = mDb.insert(TABLE_TICKER, null, values);

		Cursor cursor = mDb.query(TABLE_TICKER, TickerAllColumns, KEY_TICKER_ID
				+ " = " + insertId, null, null, null, null);
		cursor.moveToFirst();
		Ticker newComment = cursorToTicker(cursor);
		cursor.close();
		return newComment;
	}

	public void deleteTicker(Ticker ticker) {
		long id = ticker.getId();
		mDb.delete(TABLE_TICKER, KEY_TICKER_ID + " = " + id, null);
	}

	public List<Ticker> getAllTickers() {
		List<Ticker> tickers = new ArrayList<Ticker>();

		Cursor cursor = mDb.query(TABLE_TICKER, TickerAllColumns, null, null,
				null, null, null);

		cursor.moveToFirst();
		while (!cursor.isAfterLast()) {
			Ticker ticker = cursorToTicker(cursor);
			tickers.add(ticker);
			cursor.moveToNext();
		}
		cursor.close();
		return tickers;
	}
	
	public Cursor getCursorToTickers() {
		Cursor cursor = mDb.query(TABLE_TICKER, TickerAllColumns, null, null,
				null, null, null);

		if (cursor != null) {
			cursor.moveToFirst();
		}
		return cursor;
	}
	
	public void populateDefaultTickers()
	{
		createTicker("ABG.P", "Abengoa");
		createTicker("ABE", "Abertis");
		createTicker("ANA", "Acciona");
		createTicker("ACX", "Acerinox");
		createTicker("ACS", "Actividades de Construccin y Servicios");
		createTicker("AMS", "Amadeus");
		createTicker("MTS", "Arcelor Mittal");
		createTicker("POP", "Banco Popular Espaol");
		createTicker("SAB", "Banco de Sabadell");
		createTicker("SAN", "Banco Santander");
		createTicker("BKT", "Bankinter");
		createTicker("BBVA", "Banco Bilbao Vizcaya Argentaria");
		createTicker("BME", "Bolsas y Mercados Espaoles");
		createTicker("CABK", "CaixaBank");
		createTicker("DIA", "Distribuidora Internacional de Alimentacin");
		createTicker("ENG", "Enags");
		createTicker("ELE", "Endesa");
		createTicker("FCC", "Fomento de Construcciones y Contratas");
		createTicker("FER", "Ferrovial");
		createTicker("GAS", "Gas Natural");
		createTicker("GRF", "Grifols");
		createTicker("IAG", "International Airlines Group");
		createTicker("IBE", "Iberdrola");
		createTicker("ITX", "Industria de Diseo Textil");
		createTicker("IDR", "Indra Sistemas");
		createTicker("JAZ", "Jazztel");
		createTicker("MAP", "MAPFRE");
		createTicker("TL5", "Mediaset Espaa Comunicacin");
		createTicker("OHL", "Obrascn Huarte Lain");
		createTicker("REE", "Red Elctrica Corporacin");
		createTicker("REP", "Repsol");
		createTicker("SYV", "Sacyr Vallehermoso");
		createTicker("TRE", "Tcnicas Reunidas");
		createTicker("TEF", "Telefnica");
		createTicker("VIS", "Viscofan");
	}

	private Ticker cursorToTicker(Cursor cursor) {
		Ticker ticker = new Ticker();
		ticker.setId(cursor.getLong(0));
		ticker.setName(cursor.getString(1));
		ticker.setCompany(cursor.getString(2));
		return ticker;
	}

	//
	// Asset methods
	//
	public Asset createAsset(String name) {
		ContentValues values = new ContentValues();
		values.put(KEY_ASSET_NAME, name);
		values.put(KEY_ASSET_W, 0);
		values.put(KEY_ASSET_MU, 0);
		values.put(KEY_ASSET_SIGMA, 0);
		
		long insertId = mDb.insert(TABLE_ASSET, null, values);

		Cursor cursor = mDb.query(TABLE_ASSET, AssetAllColumns, KEY_ASSET_ID
				+ " = " + insertId, null, null, null, null);
		cursor.moveToFirst();
		Asset newAsset = cursorToAsset(cursor);
		cursor.close();
		return newAsset;
	}

	public void deleteAsset(Asset asset) {
		long id = asset.getId();
		mDb.delete(TABLE_ASSET, KEY_ASSET_ID + " = " + id, null);
	}

	public Asset updateAsset(Asset asset, double w, double mu, double sigma) {
		ContentValues args = new ContentValues();
		long id = asset.getId();
		args.put(KEY_ASSET_ID, id);
		args.put(KEY_ASSET_NAME, asset.getName());
		args.put(KEY_ASSET_W, w);
		args.put(KEY_ASSET_MU, mu);
		args.put(KEY_ASSET_SIGMA, sigma);

		if (mDb.update(TABLE_ASSET, args, KEY_ASSET_ID + "=" + id, null) > 0) {
			Cursor cursor = mDb.query(TABLE_ASSET, AssetAllColumns,
					KEY_ASSET_ID + " = " + id, null, null, null, null);
			cursor.moveToFirst();
			asset = cursorToAsset(cursor);
		}
		return asset;
	}

	private Asset cursorToAsset(Cursor cursor) {
		Asset asset = new Asset();
		if (!cursor.isNull(0))
			asset.setId(cursor.getLong(0));
		else
			asset.setId(0);
		
		if (!cursor.isNull(1))
			asset.setName(cursor.getString(1));
		else
			asset.setName("None");
		
		if (!cursor.isNull(2))
			asset.setW(cursor.getDouble(2));
		else
			asset.setW(0);
		
		if (!cursor.isNull(3))
			asset.setMu(cursor.getDouble(3));
		else
			asset.setMu(0);
		
		if (!cursor.isNull(4))
			asset.setSigma(cursor.getDouble(4));
		else
			asset.setSigma(0);
		return asset;
	}

	//
	// Portfolio methods
	//
	public Portfolio createPortfolio(String name) {
		ContentValues values = new ContentValues();
		values.put(KEY_PORTFOLIO_NAME, name);
		long insertId = mDb.insert(TABLE_PORTFOLIO, null, values);

		Cursor cursor = mDb.query(TABLE_PORTFOLIO, PortfolioAllColumns,
				KEY_PORTFOLIO_ID + " = " + insertId, null, null, null, null);
		cursor.moveToFirst();
		Portfolio newPortfolio = cursorToPortfolio(cursor);
		cursor.close();
		return newPortfolio;
	}

	public Portfolio updatePortfolio(Portfolio portfolio, double mu,
			double sigma, double rho) {
		ContentValues args = new ContentValues();
		long id = portfolio.getId();
		args.put(KEY_PORTFOLIO_ID, id);
		args.put(KEY_PORTFOLIO_NAME, portfolio.getName());
		args.put(KEY_PORTFOLIO_MU, mu);
		args.put(KEY_PORTFOLIO_SIGMA, sigma);
		args.put(KEY_PORTFOLIO_RHO, rho);

		if (mDb.update(TABLE_PORTFOLIO, args, KEY_PORTFOLIO_ID + "=" + id, null) > 0) {
			Cursor cursor = mDb.query(TABLE_PORTFOLIO, PortfolioAllColumns,
					KEY_PORTFOLIO_ID + " = " + id, null, null, null, null);
			cursor.moveToFirst();
			portfolio = cursorToPortfolio(cursor);
		}
		return portfolio;
	}

	public void deletePortfolio(Portfolio portfolio) {
		long id = portfolio.getId();
		mDb.delete(TABLE_PORTFOLIO_TO_ASSET, KEY_PORTFOLIO_TO_ASSET_PORTFOLIO
				+ " = " + id, null);
		mDb.delete(TABLE_PORTFOLIO, KEY_PORTFOLIO_ID + " = " + id, null);
	}

	public Portfolio addAssetToPortfolio(Asset asset, Portfolio portfolio) {
		long portfolioId = portfolio.getId();
		long assetId = asset.getId();

		ContentValues values = new ContentValues();
		values.put(KEY_PORTFOLIO_TO_ASSET_PORTFOLIO, portfolioId);
		values.put(KEY_PORTFOLIO_TO_ASSET_ASSET, assetId);

		if (mDb.insert(TABLE_PORTFOLIO_TO_ASSET, null, values) > 0) {
			portfolio.addAsset(asset);
		}
		return portfolio;
	}

	public Portfolio removeAssetFromPortfolio(Asset asset, Portfolio portfolio) {
		long portfolioId = portfolio.getId();
		long assetId = asset.getId();

		mDb.delete(TABLE_PORTFOLIO_TO_ASSET, KEY_PORTFOLIO_TO_ASSET_PORTFOLIO
				+ " = " + portfolioId + "and " + KEY_PORTFOLIO_TO_ASSET_ASSET
				+ " = " + assetId, null);
		portfolio.removeAsset(asset);
		return portfolio;
	}
	
	public List<Portfolio> getAllPortfolios() {
		List<Portfolio> portfolios = new ArrayList<Portfolio>();

		Cursor cursor = mDb.query(TABLE_PORTFOLIO, PortfolioAllColumns, null, null,
				null, null, null);

		cursor.moveToFirst();
		while (!cursor.isAfterLast()) {
			Portfolio portfolio = cursorToPortfolio(cursor);
			portfolios.add(portfolio);
			cursor.moveToNext();
		}
		cursor.close();
		return portfolios;
	}

	private Portfolio cursorToPortfolio(Cursor cursor) {
		Portfolio portfolio = new Portfolio();
		portfolio.setId(cursor.getLong(0));
		portfolio.setName(cursor.getString(1));
		portfolio.setMu(cursor.getDouble(2));
		portfolio.setSigma(cursor.getDouble(3));
		portfolio.setRho(cursor.getDouble(4));

		Cursor cursorAssets = mDb.query(TABLE_PORTFOLIO_TO_ASSET,
				PortfolioToAssetAllColumns, KEY_PORTFOLIO_ID + " = "
						+ portfolio.getId(), null, null, null, null);

		cursorAssets.moveToFirst();
		while (!cursorAssets.isAfterLast()) {
			Asset asset = cursorToAsset(cursorAssets);
			portfolio.addAsset(asset);
			cursorAssets.moveToNext();
		}

		cursorAssets.close();
		return portfolio;
	}
}
